
#include <tcl.h>
#include <tk.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "video.h"
#include "window.h"

static Tk_ConfigSpec configSpecs[] = {
    {TK_CONFIG_STRING, "-dither", "dither", "Dither",
        "ordered2", Tk_Offset(WindowConfig, ditherstring), 0},
    {TK_CONFIG_INT, "-l_range", "l_range", "L_Range",
        "8", Tk_Offset(WindowConfig, l_range), 0},
    {TK_CONFIG_INT, "-cr_range", "cr_range", "CR_Range",
        "4", Tk_Offset(WindowConfig, cr_range), 0},
    {TK_CONFIG_INT, "-cb_range", "cb_range", "CB_Range",
        "4", Tk_Offset(WindowConfig, cb_range), 0},
    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
        (char *) NULL, 0, 0}
};

/* -------------------------------------- */
/*  Old file handler...change this        */
/* -------------------------------------- */


#define BUF_LENGTH 80000

extern void rewind_video(WindowConfig *cfg,FILE *file);
extern void pause_video(WindowConfig *cfg);
extern void refresh_display(WindowConfig *cfg);
extern void single_frame(WindowConfig *cfg);
extern void run_frames(WindowConfig *cfg);


int ConfigureWindow(Tcl_Interp *interp, WindowConfig *cfg,int argc,char *argv[])
{
int ditherType;

  if (Tk_ConfigureWidget(interp,cfg->window,configSpecs,argc,argv,(char *)cfg,0) != TCL_OK) {
    return(TCL_ERROR);
  }
  if (strcmp(cfg->ditherstring, "hybrid") == 0) {
	ditherType = HYBRID_DITHER;
  } else if (strcmp(cfg->ditherstring, "hybrid2") == 0) {
	ditherType = HYBRID2_DITHER;
  } else if (strcmp(cfg->ditherstring, "fs4") == 0) {
	ditherType = FS4_DITHER;
  } else if (strcmp(cfg->ditherstring, "fs2") == 0) {
	ditherType = FS2_DITHER;
  } else if (strcmp(cfg->ditherstring, "fs2fast") == 0) {
	ditherType = FS2FAST_DITHER;
  } else if (strcmp(cfg->ditherstring, "hybrid2") == 0) {
	ditherType = HYBRID2_DITHER;
  } else if (strcmp(cfg->ditherstring, "2x2") == 0) {
	ditherType = Twox2_DITHER;
  } else if (strcmp(cfg->ditherstring, "gray") == 0) {
	ditherType = GRAY_DITHER;
  } else if (strcmp(cfg->ditherstring, "color") == 0) {
	ditherType = FULL_COLOR_DITHER;
  } else if (strcmp(cfg->ditherstring, "none") == 0) {
	ditherType = NO_DITHER;
  } else if (strcmp(cfg->ditherstring, "ordered") == 0) {
	ditherType = ORDERED_DITHER;
  } else if (strcmp(cfg->ditherstring, "ordered2") == 0) {
	ditherType = ORDERED2_DITHER;
  } else if (strcmp(cfg->ditherstring, "mbordered") == 0) {
	ditherType = MBORDERED_DITHER;
  } else if (strcmp(cfg->ditherstring, "mono") == 0) {
	ditherType = MONO_DITHER;
  } else if (strcmp(cfg->ditherstring, "threshold") == 0) {
	ditherType = MONO_THRESHOLD;
  } else {
    Tcl_AppendResult(interp,"Illegal dither option",(char *)NULL);
    return(TCL_ERROR);
  }
  cfg->ditherType = ditherType;

  if (cfg->l_range < 1) {
    Tcl_AppendResult(interp,"Illegal luminance range value",(char *)NULL);
    return(TCL_ERROR);
  }
  if (cfg->cr_range < 1) {
    Tcl_AppendResult(interp,"Illegal cr range value",(char *)NULL);
    return(TCL_ERROR);
  }
  if (cfg->cb_range < 1) {
    Tcl_AppendResult(interp,"Illegal cb range value",(char *)NULL);
    return(TCL_ERROR);
  }

  Tk_SetWindowBackground(cfg->window,BlackPixelOfScreen(Tk_Screen(cfg->window)));
  Tk_GeometryRequest(cfg->window,300,300);

  return(TCL_OK);
}

int create_ports(WindowConfig *cfg,int *in,int *out)
{
int file[2];
  if (socketpair(AF_UNIX,SOCK_STREAM,0,file) == 0) {
    fprintf(stderr,"Opened passageway to player\n");
    *in = file[0];
    *out = file[1];
    return(1);
  } else {
   fprintf(stderr,"Could not open the socketpair\n");
   return(1);
  }
  return(0);
}

void play_mpegvideo(WindowConfig *cfg)
{
int out,in;
  if (!create_ports(cfg,&in,&out)) return;
  switch(cfg->pid = fork()) {
    case 0: /* In child */
      close(in);
      cfg->comms = out;
      setup_globals(cfg);
      init_tables();
      init_display(cfg);
      process_commands(cfg);
      exit(1);  /* Not reached */
    case -1: /* Error */
      fprintf(stderr,"Could not start child process\n");
      break;
    default: /* Parent */
      close(out);
      cfg->comms = in;
      fprintf(stderr,"Child process started\n");
      fclose(cfg->input);
      break;
  }
}

void send_message(WindowConfig *cfg,char message)
{
  write(cfg->comms,&message,1);
}

void videoExposed(ClientData clientdata)
{
WindowConfig *cfg = (WindowConfig *)clientdata;

  if (cfg->first == 0) { /* First time called, so initialize first */
    cfg->first=1;
    play_mpegvideo(cfg);
  } else {
    send_message(cfg,COMMS_REFRESH);
  }
}

void DestroyVideo(ClientData clientdata)
{
WindowConfig *cfg = (WindowConfig *)clientdata;
  Tk_FreeOptions(configSpecs,(char *)cfg,cfg->display,0);
  free(cfg);
}

void windowEvent(ClientData clientdata,XEvent *eventPtr)
{
WindowConfig *cfg = (WindowConfig *)clientdata;

  if (eventPtr->type == Expose) {
    if ((eventPtr->xexpose.count == 0)) {
      Tk_DoWhenIdle(videoExposed,clientdata);
    }
  } else if (eventPtr->type == DestroyNotify) {
    Tcl_DeleteCommand(cfg->interp,Tk_PathName(cfg->window));
    send_message(cfg,COMMS_EXIT);
    Tk_EventuallyFree((ClientData)cfg,DestroyVideo);
  }
}

int videointercmd(ClientData clientData,Tcl_Interp *interp,int argc, char *argv[])
{
WindowConfig *cfg = (WindowConfig *)clientData;
  if (argc != 2) {
    Tcl_AppendResult(interp,"wrong # args: should be \"pathName interface_command\"",(char *)NULL);
    return(TCL_ERROR);
  }
  if (strcmp(argv[1],"play") == 0)        send_message(cfg,COMMS_PLAY);
  else if (strcmp(argv[1],"pause") == 0)  send_message(cfg,COMMS_PAUSE);
  else if (strcmp(argv[1],"step") == 0)   send_message(cfg,COMMS_STEP);
  else if (strcmp(argv[1],"rewind") == 0) send_message(cfg,COMMS_REWIND);
  else if (strcmp(argv[1],"setloop") == 0)   {
    send_message(cfg,COMMS_SETLOOP);
    cfg->loop = 1;
  } else if (strcmp(argv[1],"clearloop") == 0) {
    send_message(cfg,COMMS_CLEARLOOP);
    cfg->loop = 0;
  } else {
    Tcl_AppendResult(interp,"unknown command",(char *)NULL);
    return(TCL_ERROR);
  }
  return(TCL_OK);
}

int videoframecmd(ClientData clientData,Tcl_Interp *interp,int argc, char *argv[])
{
Tk_Window main = (Tk_Window) clientData;
Tk_Window window;
WindowConfig *cfg;
FILE *input;

  if (argc < 3) {
    Tcl_AppendResult(interp,"wrong # args: should be \"",argv[0], " pathName filename ?options?\"",(char *)NULL);
    return(TCL_ERROR);
  }
  if ((input = fopen(argv[2],"r")) == NULL) {
    char msg[128];
    sprintf(msg, "could not open input file : %s", argv[2]);
    Tcl_AppendResult(interp,msg,(char *)NULL);
    return(TCL_ERROR);
  }

  if ((window = Tk_CreateWindowFromPath(interp,main,argv[1],(char *)NULL)) == NULL) {
    return(TCL_ERROR);
  }
  Tk_SetClass(window,"videoplayer");

  cfg = (WindowConfig *) malloc(sizeof (WindowConfig));
  cfg->window  = window;
  cfg->interp  = interp;
  cfg->display = Tk_Display(window);
  cfg->refresh = 0;
  cfg->first   = 0;
  cfg->currentframe = 0;
  cfg->at_end = 0;
  cfg->quietFlag = 1;
  cfg->shmemFlag = 1;
  cfg->input = input;
  cfg->loop=0;

  if (ConfigureWindow(interp,cfg,argc-3,argv+3) != TCL_OK) {
    Tk_DestroyWindow(window);
    return(TCL_ERROR);
  }

  Tk_CreateEventHandler(window,ExposureMask|StructureNotifyMask,windowEvent,(ClientData) cfg);

  Tcl_CreateCommand(interp,Tk_PathName(cfg->window),(Tcl_CmdProc *)videointercmd,(ClientData *)cfg,(Tcl_CmdDeleteProc *)NULL);
  interp->result = Tk_PathName(window);

  return(TCL_OK);
}

